@@ -1,4 +1,4 @@
-/* $adsuck: adsuck.c,v 1.33 2009/03/08 23:48:10 marco Exp $ */
+/* $adsuck: adsuck.c,v 1.36 2009/03/26 03:06:28 marco Exp $ */
 /*
  * Copyright (c) 2009 Marco Peereboom <marco@peereboom.us>
  *
@@ -45,11 +45,11 @@
 
 #include "adsuck.h"
 
-#define MAXLINE		(128)
+#define MAXLINE		(256)
 #define INBUF_SIZE	(4096)
 #define LOCALIP		"127.0.0.1"
 #define ADSUCK_USER	"_adsuck"
-#define VERSION		"1.4"
+#define VERSION		"1.6"
 
 int			entries;
 int			verbose;
@@ -57,15 +57,17 @@ int			debug;
 int			debugsyslog;
 
 /* socket */
-int			sock;
+int			so;
 struct sockaddr		paddr;
 socklen_t		plen = (socklen_t) sizeof(paddr);
 
 /* resolver */
-ldns_resolver		*res;
+ldns_resolver		*resolver;
 char			*resolv_conf;
 char			*domainname;
 char			*regexfile;
+
+/* signals */
 volatile sig_atomic_t   newresolv;
 volatile sig_atomic_t   stop;
 volatile sig_atomic_t   reread;
@@ -78,8 +80,7 @@ struct regexnode {
 	char				*rname;
 };
 
-SIMPLEQ_HEAD(regexhead, regexnode);
-struct regexhead	rh;
+SIMPLEQ_HEAD(regexhead, regexnode) rh = SIMPLEQ_HEAD_INITIALIZER(rh);
 
 struct hostnode {
 	RB_ENTRY(hostnode)	hostentry;
@@ -248,34 +249,6 @@ addhosts(char *filename)
 	fclose(f);
 }
 
-int
-rereadhosts(int argc, char *argv[])
-{
-	struct hostnode		*n, *nxt;
-
-	if (!RB_EMPTY(&hosthead)) {
-		log_info("rereading blacklist entries");
-		for (n = RB_MIN(hosttree, &hosthead); n != NULL; n = nxt) {
-			nxt = RB_NEXT(hosttree, &hosthead, n);
-			RB_REMOVE(hosttree, &hosthead, n);
-			free(n);
-			entries--;
-		}
-	}
-
-	while (argc) {
-		log_info("adding %s", argv[0]);
-
-		addhosts(argv[0]);
-		argc--;
-		argv++;
-	}
-
-	log_info("total entries: %d", entries);
-
-	return (0);
-}
-
 int
 udp_bind(int sock, u_int16_t port, char *my_address)
 {
@@ -323,7 +296,10 @@ hostnamefrompkt(ldns_pkt *pkt, ldns_rr **qrr)
 
 	if (found) {
 		rawname[i] = '\0';
-		asprintf(&name, "%s", rawname);
+		if (asprintf(&name, "%s", rawname) == -1) {
+			name = NULL;
+			goto freeraw;
+		}
 		if (qrr)
 			*qrr = query_rr;
 	}
@@ -431,7 +407,7 @@ spoofquery(struct hostnode *hn, ldns_rr *query_rr, u_int16_t id)
 			logpacket(answer_pkt);
 		}
 
-		if (sendto(sock, outbuf, answer_size, 0, &paddr, plen) == -1)
+		if (sendto(so, outbuf, answer_size, 0, &paddr, plen) == -1)
 			log_warn("spoofquery sendto");
 		else {
 			rv = 0;
@@ -470,6 +446,7 @@ forwardquery(char *hostname, ldns_rr *query_rr, u_int16_t id)
 	ldns_rr_class		clas;
 	ldns_status		status;
 	int			rv = 1, child = 0;
+	struct hostnode		hn;
 
 	switch (fork()) {
 	case -1:
@@ -490,9 +467,15 @@ forwardquery(char *hostname, ldns_rr *query_rr, u_int16_t id)
 	}
 	type = ldns_rr_get_type(query_rr);
 	clas = ldns_rr_get_class(query_rr);
-	respkt = ldns_resolver_query(res, qname, type, clas, qflags);
+	respkt = ldns_resolver_query(resolver, qname, type, clas, qflags);
 	if (respkt == NULL) {
-		log_warnx("forwardquery: no respkt");
+		/* dns query failed so lets spoof it instead of timing out */
+		log_warnx("forwardquery: query failed, spoofing response");
+
+		/* XXX make this tunable? */
+		hn.ipaddr = NULL;
+		hn.hostname = hostname;
+		spoofquery(&hn, query_rr, id);
 		goto unwind;
 	}
 	if (debug) {
@@ -506,7 +489,7 @@ forwardquery(char *hostname, ldns_rr *query_rr, u_int16_t id)
 		log_warnx("can't create answer: %s",
 		    ldns_get_errorstr_by_id(status));
 	else {
-		if (sendto(sock, outbuf, answer_size, 0, &paddr, plen) == -1)
+		if (sendto(so, outbuf, answer_size, 0, &paddr, plen) == -1)
 			log_warn("forwardquery sendto");
 		else {
 			rv = 0;
@@ -537,24 +520,25 @@ setupresolver(void)
 	char			*action = "using", *es;
 	char			buf[128];
 	ldns_rdf		*dn;
-	int			i;
+	size_t			i;
 
-	if (res) {
-		ldns_resolver_free(res);
+	if (resolver) {
+		ldns_resolver_free(resolver);
 		free(domainname); /* XXX is this ok for ldns? */
-		res = NULL;
+		resolver = NULL;
 		domainname = NULL;
 		action = "rereading";
 	}
 
-	status = ldns_resolver_new_frm_file(&res, resolv_conf);
+	status = ldns_resolver_new_frm_file(&resolver, resolv_conf);
 	if (status != LDNS_STATUS_OK) {
-		asprintf(&es, "bad resolv.conf file: %s",
-			ldns_get_errorstr_by_id(status));
+		if (asprintf(&es, "bad resolv.conf file: %s",
+		    ldns_get_errorstr_by_id(status)) == -1)
+			fatal("setupresolver");
 		fatalx(es);
 	}
 
-	dn = ldns_resolver_domain(res);
+	dn = ldns_resolver_domain(resolver);
 	if (dn == NULL) {
 		domainname = NULL;
 		if (gethostname(buf, sizeof buf) == -1) {
@@ -567,7 +551,8 @@ setupresolver(void)
 
 			if (buf[i] == '.' && strlen(buf) > 1) {
 				i++;
-				asprintf(&domainname, "%s", &buf[i]);
+				if (asprintf(&domainname, "%s", &buf[i]) == -1)
+					fatal("setupresolver");
 			}
 		}
 	} else {
@@ -603,6 +588,43 @@ freeregex(void)
 	SIMPLEQ_INIT(&rh);
 }
 
+void
+freerb(void)
+{
+	struct hostnode		*n, *nxt;
+
+	if (RB_EMPTY(&hosthead))
+		return;
+
+	for (n = RB_MIN(hosttree, &hosthead); n != NULL; n = nxt) {
+		nxt = RB_NEXT(hosttree, &hosthead, n);
+		RB_REMOVE(hosttree, &hosthead, n);
+		free(n);
+		entries--;
+	}
+	RB_INIT(&hosthead);
+}
+
+int
+rereadhosts(int argc, char *argv[])
+{
+	freerb();
+
+	log_info("rereading blacklist entries");
+
+	while (argc) {
+		log_info("adding %s", argv[0]);
+
+		addhosts(argv[0]);
+		argc--;
+		argv++;
+	}
+
+	log_info("total entries: %d", entries);
+
+	return (0);
+}
+
 int
 setupregex(void)
 {
@@ -611,8 +633,7 @@ setupregex(void)
 	int			i = 0, rv;
 	struct regexnode	*n;
 
-	if (!SIMPLEQ_EMPTY(&rh))
-		freeregex();
+	freeregex();
 
 	if (regexfile == NULL)
 		return (0);
@@ -782,10 +803,10 @@ main(int argc, char *argv[])
 	if ((pw = getpwnam(user)) == NULL)
 		errx(1, "unknown user %s", user);
 
-	sock =  socket(AF_INET, SOCK_DGRAM, 0);
-	if (sock == -1)
+	so = socket(AF_INET, SOCK_DGRAM, 0);
+	if (so == -1)
 		err(1, "can't open socket");
-	if (udp_bind(sock, port, listen_addr))
+	if (udp_bind(so, port, listen_addr))
 		err(1, "can't udp bind");
 
 	/* daemonize */
@@ -830,11 +851,10 @@ main(int argc, char *argv[])
 	rereadhosts(argc, argv);
 
 	/* regex */
-	SIMPLEQ_INIT(&rh);
 	rcount = setupregex();
 
 	while (!stop) {
-		nb = recvfrom(sock, inbuf, INBUF_SIZE, 0, &paddr, &plen);
+		nb = recvfrom(so, inbuf, INBUF_SIZE, 0, &paddr, &plen);
 		if (nb == -1) {
 			if (errno == EINTR || errno == EAGAIN) {
 				dosignals(argc, argv);
@@ -857,6 +877,8 @@ main(int argc, char *argv[])
 
 		bzero(&hostn, sizeof hostn);
 		hostn.hostname = hostnamefrompkt(query_pkt, &query_rr);
+		if (hostn.hostname == NULL)
+			continue; /* maybe this should be fatal */
 		id = ldns_pkt_id(query_pkt);
 		if (domainname &&
 		    (s = strstr(hostn.hostname, domainname)) != NULL) {
@@ -865,7 +887,8 @@ main(int argc, char *argv[])
 			 * without domain name; this is to work around
 			 * software that tries to be smart about domain names
 			 */
-			asprintf(&h.hostname, "%s", hostn.hostname);
+			if (asprintf(&h.hostname, "%s", hostn.hostname) == -1)
+				fatal("hostname");
 			h.hostname[s - hostn.hostname - 1] = '\0';
 			if (runregex(h.hostname) == 0)
 				spoofquery(&hostn, query_rr, id);
@@ -889,6 +912,7 @@ main(int argc, char *argv[])
 	}
 
 	freeregex();
+	freerb();
 
 	log_info("exiting");
 
